@@ -1,5 +1,5 @@
 /* Definitions for computing resource usage of specific insns.
-   Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.
+   Copyright (C) 1999, 2000, 2001, 2003 Free Software Foundation, Inc.
 
 This file is part of GCC.
 
@@ -504,7 +504,12 @@ find_dead_or_set_registers (target, res, jump_target, jump_count, set, needed)
 	  if (jump_count++ < 10)
 	    {
 	      if (any_uncondjump_p (this_jump_insn)
-		  || GET_CODE (PATTERN (this_jump_insn)) == RETURN)
+		  || GET_CODE (PATTERN (this_jump_insn)) == RETURN
+		  /* BEGIN AXIS LOCAL */
+		  || (GET_CODE (PATTERN (this_jump_insn)) == PARALLEL
+		      && (GET_CODE (XVECEXP (PATTERN (this_jump_insn), 0, 0))
+			  == RETURN)))
+		  /* END AXIS LOCAL */
 		{
 		  next = JUMP_LABEL (this_jump_insn);
 		  if (jump_insn == 0)
