@@ -1,10 +1,19 @@
 /* Call ctors and dtors from shared libraries, if any.
-   Copyright (C) 1999, 2000 Axis Communications.  */
+   Copyright (C) 1999, 2000, 2003 Axis Communications.  */
 
 typedef void (*vfnp) (void);
 
-vfnp * const _Ctors asm(".$global.lib.ctors") = {0};
-vfnp * const _Dtors asm(".$global.lib.dtors") = {0};
+/* FIXME: gcc-3.2.1 "optimized" away the guts of the _Libctors and
+   _Libdtors into empty functions when the definition was here as well.
+   (For that version, attribute section isn't supported for a.out.)
+   Moving the definitions to after the function is a
+   gcc-version-specific trick, which will fail with a later gcc version.
+   Until then, implement the section attribute for a.out (.text, .data
+   only).  Alternatively, make gcc not look into asm-declared consts.
+   Still alternatively, emit the definitions as asm.  They absolutely
+   have to end up in the .text section.  */
+extern vfnp * const _Ctors asm(".$global.lib.ctors");
+extern vfnp * const _Dtors asm(".$global.lib.dtors");
 
 /* We better provide weak empty ctor and dtor lists, since they are not
    created if the main program does not have ctor/dtors.
@@ -48,3 +57,6 @@ void _Libdtors(void)
       (**dtors++) ();
     }
 }
+
+vfnp * const _Ctors asm(".$global.lib.ctors") = {0};
+vfnp * const _Dtors asm(".$global.lib.dtors") = {0};
